# #!/usr/bin/env python
# # -*- coding: utf-8 -*-

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from SuPyMode.supermode import SuperMode

from SuPyMode.representation.base import InheritFromSuperMode, BaseSingleModePlot
from MPSPlots.render2D import SceneList, Axis


class Index(InheritFromSuperMode, BaseSingleModePlot):
    """
    Represents the effective refractive index of a mode derived from a supermode in optical fiber simulations.

    This class extends from `InheritFromSuperMode` for accessing supermode-related data and `BaseSingleModePlot`
    for plotting functionalities tailored to visualize the effective refractive index.

    Class Attributes:
        plot_style (dict): A dictionary defining the default style settings for plots generated by this class.
    """

    plot_style = dict(
        show_legend=True,
        x_label='Inverse taper ratio',
        y_label='Effective refraction index',
        y_scale="linear",
        y_limits=[1.44, 1.455],
        line_width=2
    )

    def __init__(self, parent_supermode: SuperMode):
        """
        Initializes an Index object with a reference to a parent supermode.

        Args:
            parent_supermode (SuperMode): The parent supermode object that provides the base mode data.
        """
        self.parent_supermode = parent_supermode

    @property
    def data(self):
        return self.parent_supermode.binded_supermode.get_index()

    def render_on_ax(self, ax: Axis) -> None:
        """
        Renders the effective refractive index data as a line plot on the provided Axis object.

        Args:
            ax (Axis): The Axis object on which to plot the effective refractive index.

        Note:
            This method utilizes the `plot_style` class attribute to define the appearance of the plot.
        """
        ax.add_line(
            x=self.itr_list,
            y=self.data,
            label=f'{self.stylized_label}'
        )

    def plot(self) -> SceneList:
        """
        Generates a plot of the effective refractive index using a SceneList.

        This method creates a single-axis plot showing the effective refractive index as a function of the inverse taper ratio,
        formatted according to the predefined plot style.

        Returns:
            SceneList: A scene list containing the plot of effective refractive index.
        """
        figure = SceneList()

        ax = figure.append_ax()

        ax.set_style(**self.plot_style)

        self.render_on_ax(ax=ax)

        return figure


# -
